public enum GridStyle {
	NONE,
	LINES,
	CROSSES,
	POLAR
}

public delegate bool DocumentForeachFunc (AbstractPlot plot);
public class Document: Object {	
	/* Private fields */
	Library		library;
	Gtk.ListStore	list;
	File?		file;
	
	/* Document data */
	public double min_x         { get; set; default = -10.0; }
	public double max_x         { get; set; default =  10.0; }
	public double shift_y       { get; set; default =   0.0; }
	public bool   show_axes     { get; set; default =  true; }
	public bool   show_labels   { get; set; default =  true; }
	public bool   trigonometric { get; set; default = false; }
	public GridStyle grid_style { get; set; default = GridStyle.LINES; }
	
	/* Signals */
	public signal void changed ();
	
	/* Methods */
	public Document () { base_construct (); }
	public Library get_library () { return library; }
	public Gtk.ListStore get_list () { return list; }
	public File? get_file () { return file; }
	
	public void list_set_plot (Gtk.TreeIter iter, AbstractPlot plot)
	{
		string text = "(unknown)";
		if (plot is SingleExpressionPlot) {
			text = (plot as SingleExpressionPlot).expression;
		} else if (plot is CoupleExpressionPlot) {
			var ce_plot = plot as CoupleExpressionPlot;
			text = ce_plot.expression_x + "\n" + ce_plot.expression_y;
		}
		plot.notify.connect ((s, p) => { changed (); });
		
		list.set (iter, 0, plot.visible,
				1, text,
				2, plot,
				-1);
	}
	
	public void list_set_visibility (Gtk.TreeIter iter, bool visible)
	{
		list_get_plot (iter).visible = visible;
		list.set (iter, 0, visible, -1);
	}
	
	public void list_set_text (Gtk.TreeIter iter, string text)
	{
		list.set (iter, 1, text, -1);
	}
	
	public AbstractPlot list_get_plot (Gtk.TreeIter iter)
	{
		Value value;
		list.get_value (iter, 2, out value);
		return value as AbstractPlot;
	}
	
	public void list_foreach (DocumentForeachFunc func)
	{
		list.foreach ((model, path, iter) => {
			return func (list_get_plot (iter));
		});
	}

	private void base_construct ()
	{
		library = new Library ();
		library.register (Function.from_string ("sqrt(x)=x^0.5"));
		library.register (new FunctionSin ());
		library.register (new FunctionCos ());
		library.register (new FunctionTan ());
		library.register (new FunctionAbs ());
		library.register (new FunctionExp ());
		library.register (new FunctionLog ());
		library.register (new FunctionSinh ());
		library.register (new FunctionCosh ());
		library.register (new FunctionTanh ());
		library.register (new FunctionAcos ());
		library.register (new FunctionAsin ());
		library.register (new FunctionAtan ());
		
		list = new Gtk.ListStore (3,
					  typeof (bool), /* 0: VISIBILITY */
					  typeof (string), /* 1: TEXT */
					  typeof (AbstractPlot) /* 2: PLOT */
					  );
		list.row_deleted.connect ((t, p) => { changed (); });
		list.row_inserted.connect ((t, p) => { changed (); });
		list.rows_reordered.connect ((t, p, i, a3) => { changed (); });
		notify.connect ( (s, p) => { changed (); });
	}
	
	[CCode(has_type_id = false)]
	private struct ParserData {
		Document document;
		AbstractPlot? plot;
		Library library;
		string? property;
		Value prop_string;
		Value prop_value;
		Object? obj;
	}
	
	public Document.from_file (File file) throws Error, MarkupError
	{
		MarkupParser parser = { /* start element */
		(context, element, attr_name, attr_value) => {
			ParserData* data = (ParserData*) this;
			
			unowned SList <string> element_stack = context.get_element_stack ();
			unowned string? parent = null;
			if (element_stack.next != null)
				parent = element_stack.next.data;
			
			int line, column;
			context.get_position (out line, out column);
			string err = _("Error on line %d char %d: ").printf (line, column);
			
			switch (element) {
			case "vplotdoc":
				if (parent != null) {
					unowned string msg = _("Unexpected 'vplotdoc' element");
					throw new MarkupError.UNKNOWN_ELEMENT (err + msg);
				}
				if (attr_name.length != 1 || attr_name[0] != "version") {
					unowned string msg = _("Missing 'version' attribute");
					throw new MarkupError.MISSING_ATTRIBUTE (err + msg);
				}
				break;
			case "plots":
				if (parent != "vplotdoc") {
					unowned string msg = _("Unexpected 'plots' element");
					throw new MarkupError.UNKNOWN_ELEMENT (err + msg);
				}
				if (attr_name.length != 0) {
					unowned string msg = _("Unexpected attribute for 'plots' element");
					throw new MarkupError.UNKNOWN_ATTRIBUTE (err + msg);
				}
				break;
			case "plot":
				if (parent != "plots") {
					unowned string msg = _("Unexpected 'plot' element");
					throw new MarkupError.UNKNOWN_ELEMENT (err + msg);
				}
				if (attr_name.length != 1 || attr_name[0] != "type") {
					unowned string msg = _("Missing 'type' attribute for 'plot' element");
					throw new MarkupError.UNKNOWN_ATTRIBUTE (err + msg);
				}
				
				switch (attr_value[0]) {
				case "CartesianPlot":
					data->plot = new CartesianPlot (data->library);
					break;
				case "ParametricPlot":
					data->plot = new ParametricPlot (data->library);
					break;
				case "PolarPlot":
					data->plot = new PolarPlot (data->library);
					break;
				default:
					warning (_("Unknown plot type: %s"), attr_value[0]);
				}
				break;
			default:
				switch (parent) {
				case "vplotdoc":
					data->obj = data->document;
					break;
				case "plot":
					data->obj = data->plot;
					break;
				default:
					unowned string msg = _("Unknown element: %s");
					throw new MarkupError.UNKNOWN_ELEMENT (err + msg.printf (element));
				}
				
				weak ObjectClass klass = data->obj.get_class ();
				ParamSpec? pspec = klass.find_property (element);
				if (pspec == null) {
					unowned string msg = _("Unknown element: %s");
					throw new MarkupError.UNKNOWN_ELEMENT (err + msg.printf (element));
				}
				if (attr_name.length == 0) {
					unowned string msg = _("Missing 'type' attribute");
					throw new MarkupError.MISSING_ATTRIBUTE (err + msg);
				}
				if (attr_name.length > 1 || attr_name[0] != "type") {
					unowned string msg = _("Too much attributes");
					throw new MarkupError.UNKNOWN_ATTRIBUTE (err + msg);
				}
				if (attr_value[0] != pspec.value_type.name ()) {
					unowned string msg = _("Type mismatch");
					throw new MarkupError.INVALID_CONTENT (err + msg);
				}
				
				data->prop_value.unset ();
				data->prop_value.init (pspec.value_type);
				data->property = element;
			}
		},
		(context, element) => { /* end element */
			ParserData* data = (ParserData*) this;
			if (element == "plot") {
				Gtk.TreeIter iter;
				data->document.list.append (out iter);
				data->document.list_set_plot (iter, data->plot);
				data->plot = null;
			}
		},
		(context, text, text_len) => { /* text */
			ParserData* data = (ParserData*) this;
			if (data->obj == null)
				return;
			
			char* text_copy = Memory.dup (text, (uint) text_len + 1);
			text_copy[text_len] = 0;
			data->prop_string.take_string ((string) (owned) text_copy);
			if (data->prop_string.transform (ref data->prop_value)) {
				data->obj.set_property (data->property, data->prop_value);
			} else {
				warning (_("Can't convert string to %s"), data->prop_value.type_name ());
			}
			data->obj = null;
		},
			null,
			null
		};
		
		base_construct ();
		ParserData data = ParserData () {
			document = this,
			plot = null,
			library = this.library,
			property = null,
			prop_string = Value (typeof (string)),
			prop_value = Value (typeof (int)),
			obj = null
		};
		
		/* Reset locale */
		Intl.setlocale (LocaleCategory.ALL, "C");
		
		try {
			MarkupParseContext context = new MarkupParseContext (parser, 0, &data, null);
			FileInputStream stream = file.read (null);
			char[] buffer = new char[4096];
			ssize_t buflen = -1;
		
			while ((buflen = stream.read (buffer, buffer.length, null)) != 0)
				context.parse ((string) buffer, buflen);
			stream.close (null);
			this.file = file;
		} finally {
			/* Restore locale */
			Gtk.set_locale ();
		}
	}
	
	private static void write_object_to_xml_stream (Object obj, FileOutputStream stream, string prefix = "") throws Error
	{
		weak ObjectClass klass = obj.get_class ();
		weak ParamSpec[] properties = klass.list_properties ();
		
		foreach (ParamSpec prop in properties) {
			Value prop_value = Value (prop.value_type);
			Value prop_string = Value (typeof (string));
			
			obj.get_property (prop.name, ref prop_value);
			if (prop_value.transform (ref prop_string)) {
				string buffer = Markup.printf_escaped ("%s<%s type=\"%s\">%s</%s>\n",
									prefix,
									prop.name,
									prop.value_type.name (),
									prop_string.get_string (),
									prop.name);
				stream.write (buffer, buffer.size (), null);
			} else {
				warning (_("Can't convert %s to string"), prop.value_type.name ());
			}
		}
		
		free (properties);
	}
	
	public void save_to_file (File file) throws Error
	{
		/* Reset locale */
		Intl.setlocale (LocaleCategory.ALL, "C");
		
		try {
			FileOutputStream stream = file.replace (null, false, 0, null);
		
			/* Header */
			string buffer = "<!DOCTYPE vplotdoc>\n<vplotdoc version=\"1\">\n";
			stream.write (buffer, buffer.size (), null);
		
			/* Body */
			write_object_to_xml_stream (this, stream, "\t");
			buffer = "\t<plots>\n";
			stream.write (buffer, buffer.size (), null);
			Gtk.TreeIter iter;
			list.get_iter_first (out iter);
			do {
				var plot = list_get_plot (iter);
				buffer = Markup.printf_escaped ("\t\t<plot type=\"%s\">\n",
								plot.get_type ().name ());
				stream.write (buffer, buffer.size (), null);
				write_object_to_xml_stream (plot, stream, "\t\t\t");
				buffer = "\t\t</plot>\n";
				stream.write (buffer, buffer.size (), null);
			} while (list.iter_next (ref iter));
		
			/* Footer */
			buffer = "\t</plots>\n</vplotdoc>\n";
			stream.write (buffer, buffer.size (), null);
			stream.close (null);
			
			this.file = file;
		} finally {
			/* Restore locale */
			Gtk.set_locale ();
		}
	}
}

