public class Function: FunctionIface, Object {
	/* Type declarations */
	private enum InstructionType {
		STOP,
		PUSH_ARG,
		PUSH_CONSTANT,
		ADD,
		SUBSTRACT,
		MULTIPLY,
		DIVIDE,
		POWER,
		CALL,
		INVALID
	}
	
	private struct Instruction {
		InstructionType	type;
		uint		func_id;
		uint		derivative;
		double		constant;
	}
	
	private struct ImportEntry {
		string			name;
		weak FunctionIface?	function;
	}
	
	/* Private fields */
	private Instruction*	code;
	private ImportEntry*	imports;
	private double*		stack;
	private Function?	derivative;
	private string		name;
	private bool		linking;
	
	/* Interface implementation */
	
	public unowned string get_name () { return name; }
	
	public double evaluate (double arg)
	{
		Instruction*	inst  = code;
		double*		stack = this.stack;
		
		while (inst->type != InstructionType.STOP) {
			switch (inst->type) {
			case InstructionType.PUSH_ARG:
				*(stack++) = arg;
				break;
			case InstructionType.PUSH_CONSTANT:
				*(stack++) = inst->constant;
				break;
			case InstructionType.ADD:
				*(stack-2) += *(stack-1);
				stack--;
				break;
			case InstructionType.SUBSTRACT:
				*(stack-2) -= *(stack-1);
				stack--;
				break;
			case InstructionType.MULTIPLY:
				*(stack-2) *= *(stack-1);
				stack--;
				break;
			case InstructionType.DIVIDE:
				*(stack-2) /= *(stack-1);
				stack--;
				break;
			case InstructionType.POWER:
				*(stack-2) = Math.pow (*(stack-2), *(stack-1));
				stack--;
				break;
			case InstructionType.CALL:
				FunctionIface obj = imports[inst->func_id].function;
				for (int i = 0; i < inst->derivative; i++)
					obj = obj.derivate ();
				*(stack-1) = obj.evaluate (*(stack-1));
				break;
			default:
				error ("Function: Unknown instruction\n");
				break;
			}
			inst++;
		}
		
		if (this.stack != null)
			return *this.stack;
		else
			return double.NAN;
	}
	
	public bool link_with (Library library)
	{
		/* Prevent cross references */
		if (linking) return false;
		linking = true;
		
		ImportEntry* entry = imports;
		while (entry->name != null) {
			entry->function = library.lookup (entry->name);
			if (entry->function == null) {
				linking = false;
				return false;
			}
			if (!entry->function.link_with (library)) {
				linking = false;
				return false;
			}
			entry++;
		}
		
		linking = false;
		return true;
	}
	
	public FunctionIface derivate ()
	{
		if (likely (derivative != null))
			return derivative;
		
		Function func = new Function ();
		func.name = "d_" + name;
		/* Clone imports */
		int n_imports = 0;
		ImportEntry* entry = imports;
		while (entry->name != null) {
			n_imports++;
			entry++;
		}
		func.imports = new ImportEntry[n_imports + 1];
		for (int i = 0; i <= n_imports; i++)
			func.imports[i].name = imports[i].name;
		
		/* Main */
		/* TODO */
		
		return derivative = func;
	}
	
	~Function ()
	{
		if (imports != null) {
			ImportEntry* entry = imports;
			while (entry->name != null) {
				free (entry->name);
				entry++;
			}
		}
		free (imports);
		free (code);
		free (stack);
	}
	
	private Function () {}
	
	/* Parser */
	
	private static struct ParserData {
		Function	func;
		weak string	cursor;
		unichar		argument;
		bool		dry_run;
		uint		stack_top;
		uint		code_size;
		uint		n_imports;
		uint		stack_size;
	}
	
	/* Public functions */
	public static Function? from_string (string? expression, out unowned string error = null)
	{
		ParserData data = ParserData ();
		data.func = new Function ();
		data.cursor = expression;
		
		if ((data.func.name = parse_name (data)) == null) {
			if (&error != null) error = data.cursor;
			return null;
		}	
		if (!parse_argument (data)) {
			if (&error != null) error = data.cursor;
			return null;
		}
		skip_spaces (data);
		if (data.cursor.get_char () != '=') {
			if (&error != null) error = data.cursor;
			return null;
		}
		data.cursor = data.cursor.next_char ();

		weak string expr_begin = data.cursor;
		data.dry_run = true;
		if (!parse_expr (data)) {
			if (&error != null) error = data.cursor;
			return null;
		}
		
		data.func.code = new Instruction [data.code_size + 1];
		data.func.code[data.code_size].type = InstructionType.STOP;
		data.func.stack = new double [data.stack_size];
		data.func.imports = new ImportEntry [data.n_imports + 1];
		data.func.imports[data.n_imports].name = null;
		//print ("code_size: %u\nstack_size:%u\nn_imports: %u\n", data.code_size, data.stack_size, data.n_imports);
		
		data.cursor = expr_begin;
		data.stack_top = data.code_size = data.n_imports = 0;
		data.dry_run = false;
		parse_expr (data);
		return data.func;
	}
	
	private static bool is_alnum (unichar c)
	{
		return c.isalnum () || c == '_';
	}
	
	private static bool is_alpha (unichar c)
	{
		return c.isalpha () || c == '_';
	}
	
	private static void skip_spaces (ParserData data)
	{
		while (data.cursor.get_char ().isspace ())
			data.cursor = data.cursor.next_char ();
	}
	
	private static string? parse_name (ParserData data)
	{
		skip_spaces (data);
		if (!is_alpha (data.cursor.get_char()))
			return null;
		weak string begin = data.cursor;
		data.cursor = data.cursor.next_char ();
		while (is_alnum (data.cursor.get_char()))
			data.cursor = data.cursor.next_char ();
		
		uint name_len = (uint) ((char*) data.cursor - (char*) begin);
		char* result = Memory.dup ((char*) begin, name_len + 1);
		result[name_len] = 0;
		return (string) (owned) result;
	}
	
	private static bool parse_argument (ParserData data)
	{
		skip_spaces (data);
		if (data.cursor.get_char () != '(')
			return false;
		data.cursor = data.cursor.next_char ();
		skip_spaces (data);
		data.argument = data.cursor.get_char ();
		if (!data.argument.isalpha ())
			return false;
		data.cursor = data.cursor.next_char ();
		skip_spaces (data);
		if (data.cursor.get_char () != ')')
			return false;
		data.cursor = data.cursor.next_char ();
		return true;
	}
	
	private static void push_code (ParserData data, Instruction inst)
	{
		if (!data.dry_run)
			data.func.code[data.code_size] = inst;
		if (inst.type == InstructionType.PUSH_ARG || inst.type == InstructionType.PUSH_CONSTANT)
			data.stack_top++;
		else if (inst.type != InstructionType.CALL)
			data.stack_top--;
		if (data.stack_top > data.stack_size)
			data.stack_size = data.stack_top;
		data.code_size++;
	}

	private static double parse_number (ParserData data)
	{
		double res = 0.0;
		unichar c;
		
		while ((c = data.cursor.get_char ()).isdigit ()) {
			res = res * 10.0 + c.digit_value ();
			data.cursor = data.cursor.next_char ();
		}
		
		if (data.cursor.get_char () == '.') {
			data.cursor = data.cursor.next_char ();
			double t = 0.1;
			while ((c = data.cursor.get_char ()).isdigit ()) {
				res += c.digit_value () * t;
				t /= 10.0;
				data.cursor = data.cursor.next_char ();
			}
		}
		
		c = data.cursor.get_char ();
		if (c == 'e' || c == 'E') {
			data.cursor = data.cursor.next_char ();
			c = data.cursor.get_char ();
			
			int sign = 1, power = 0;
			if (c == '+') {
				sign = 1;
				data.cursor = data.cursor.next_char ();
			} else if (c == '-') {
				sign = -1;
				data.cursor = data.cursor.next_char ();
			}
			
			while ((c = data.cursor.get_char ()).isdigit ()) {
				power = power * 10 + c.digit_value ();
				data.cursor = data.cursor.next_char ();
			}
			res = res * Math.pow (10.0, power * sign);
		}
		
		//print ("number: %lf\n", res);
		return res;
	}
	
	private static bool parse_operand (ParserData data)
	{
		//print ("parse_operand: %s\n", data.cursor);
		skip_spaces (data);
		if (data.cursor.get_char () == '(') {
			data.cursor = data.cursor.next_char ();
			if (!parse_item (data))
				return false;
			if (data.cursor.get_char () != ')')
				return false;
			data.cursor = data.cursor.next_char ();
			return true;
		}
		
		if (data.cursor.get_char ().isdigit ()) {
			push_code (data, {InstructionType.PUSH_CONSTANT, 0, 0, parse_number (data)});
			return true;
		}
		
		string? name = parse_name (data);
		if (name == null)
			return false;
		
		if (name.length == 1 && name.get_char () == data.argument) {
			push_code (data, {InstructionType.PUSH_ARG, 0, 0, 0.0});
			return true;
		}
		
		if (name == "Pi" || name == "π") {
			push_code (data, {InstructionType.PUSH_CONSTANT, 0, 0, Math.PI});
			return true;
		}
		
		skip_spaces (data);
		if (data.cursor.get_char () == '(') {
			data.cursor = data.cursor.next_char ();
			if (!parse_item (data))
				return false;
			if (data.cursor.get_char () != ')')
				return false;
			data.cursor = data.cursor.next_char ();
		} else {
			push_code (data, {InstructionType.PUSH_ARG, 0, 0, 0.0});
		}
		
		if (!data.dry_run) {
			data.func.imports[data.n_imports].name = (owned) name;
			data.func.imports[data.n_imports].function = null;
		}
		push_code (data, {InstructionType.CALL, data.n_imports, 0, 0.0});
		data.n_imports++;
		return true;
	}
	
	private static bool parse_power (ParserData data)
	{
		//print ("parse_power: %s\n", data.cursor);
		if (!parse_operand (data))
			return false;

		skip_spaces (data);
		if (data.cursor.get_char () == '^') {
			data.cursor = data.cursor.next_char ();
			if (!parse_power (data))
				return false;
			push_code (data, {InstructionType.POWER, 0, 0, 0.0});
		}
		return true;
	}
	
	private static bool parse_mult (ParserData data)
	{
		//print ("parse_mult: %s\n", data.cursor);
		if (!parse_power (data))
			return false;
		
		while (true) {
			skip_spaces (data);
			InstructionType it;
			switch (data.cursor.get_char ()) {
			case '/':
				it = InstructionType.DIVIDE;
				break;
			case '*':
				it = InstructionType.MULTIPLY;
				break;
			default:
				return true;
			}
			data.cursor = data.cursor.next_char ();
			
			if (!parse_power (data))
				return false;
			push_code (data, {it, 0, 0, 0.0});
		}
	}
	
	private static bool parse_item (ParserData data)
	{
		//print ("parse_item: %s\n", data.cursor);
		skip_spaces (data);
		bool minus = false;
		if (data.cursor.get_char () == '-') {
			minus = true;
			push_code (data, {InstructionType.PUSH_CONSTANT, 0, 0, 0.0});
			data.cursor = data.cursor.next_char ();
		}
		
		if (!parse_mult (data))
			return false;
		if (minus)
			push_code (data, {InstructionType.SUBSTRACT, 0, 0, 0.0});
		
		while (true) {
			skip_spaces (data);
			InstructionType it;
			switch (data.cursor.get_char ()) {
			case '-': 
				it = InstructionType.SUBSTRACT;
				break;
			case '+':
				it = InstructionType.ADD;
				break;
			default:
				return true;
			}
			data.cursor = data.cursor.next_char ();
			
			if (!parse_mult (data))
				return false;
			push_code (data, {it, 0, 0, 0.0});
		}
	}

	private static bool parse_expr (ParserData data)
	{
		//print ("parse_expr: %s\n", data.cursor);
		parse_item (data);
		if (data.cursor.get_char () != 0)
			return false;
		return true;
	}
}

