using GLib;
using Cairo;
using Gtk;

public class PlotArea: Gtk.DrawingArea {
	private Document document;
	private double grid_size;
	private double axis_x;
	private double axis_y;
	private double pixel_x;
	private double pixel_y;
	
	private double scale ()
	{
		return allocation.width / (document.max_x - document.min_x);
	}
	
	public PlotArea (Document document)
	{
		this.document = document;
		this.visible = true;

		Gdk.Color background_color = {0, 0xffff, 0xffff, 0xffff};
		modify_bg (StateType.NORMAL, background_color);
		add_events (Gdk.EventMask.BUTTON_PRESS_MASK   |
			    Gdk.EventMask.BUTTON_RELEASE_MASK |
			    Gdk.EventMask.POINTER_MOTION_MASK |
			    Gdk.EventMask.SCROLL_MASK);
		
		document.changed.connect ( () => {
			if (is_realized ()) {
				weak Gdk.Region region = window.get_clip_region ();
				window.invalidate_region (region, true);
			}
		});
	}
	
	private override bool button_press_event (Gdk.EventButton event)
	{
		if (event.type == Gdk.EventType.BUTTON_PRESS && event.button == 1) {
			window.set_cursor (new Gdk.Cursor (Gdk.CursorType.FLEUR));
			double begin_y = event.y;
			double begin_x = event.x;
			ulong motion_event = 0, release_event = 0;
			double k = scale();
			
			motion_event = motion_notify_event.connect ( (event) => {
				document.min_x -= (event.x - begin_x) / k;
				document.max_x -= (event.x - begin_x) / k;
				document.shift_y += event.y - begin_y;
				begin_y = event.y;
				begin_x = event.x;
				return true;
			});
			release_event = button_release_event.connect ( (event) => {
				window.set_cursor (null);
				SignalHandler.disconnect (this, motion_event);
				SignalHandler.disconnect (this, release_event);
				return true;
			});
			return true;
		}
		return false;
	}

	private override bool scroll_event (Gdk.EventScroll event)
	{
		if (event.direction == Gdk.ScrollDirection.DOWN)
			zoom_to (event.x, event.y, 2.0);
		else if (event.direction == Gdk.ScrollDirection.UP)
			zoom_to (event.x, event.y, 0.5);
		else return false;
		return true;
	}
	
	public void device_to_user (ref double x, ref double y)
	{
		x += document.min_x * scale ();
		y -= allocation.height / 2.0 + document.shift_y;
		x /= scale ();
		y /= -scale ();
	}
	
	public void device_to_user_distance (ref double x, ref double y)
	{
		x /= scale ();
		y /= scale ();
	}
	
	public static int get_precision (double eps)
	{
		double t = 1.0;
		int res = 0;
		while (t > eps) {
			t /= 10.0;
			res++;
		}
		return res;
	}
	
	public void zoom_to (double x, double y, double factor)
	{
		double view_size = document.max_x - document.min_x;
		if (view_size > 1e7 && factor > 1.0)
			return;
		if (view_size < 1e-6 && factor < 1.0)
			return;
		
		document.shift_y = (y - allocation.height / 2.0) + 
		                   (document.shift_y + allocation.height / 2.0 - y) / factor;
		device_to_user (ref x, ref y);
		document.max_x = (document.max_x - x) * factor + x;
		document.min_x = x - (x - document.min_x) * factor;
	}
	
	/*
	 * Makes grid labels readable by adjusting to
	 * 1x, 2x, 5x powers of ten.
	 */
	private static double readable_grid_size (double user_size)
	{
		double best_size = user_size;
		double min_size  = GLib.Math.pow10 (Math.floor (Math.log10 (user_size)));
		
		user_size = min_size;
		if (Math.fabs (user_size - best_size) > Math.fabs (min_size * 2 - best_size))
			user_size = min_size * 2;
		if (Math.fabs (user_size - best_size) > Math.fabs (min_size * 5 - best_size))
			user_size = min_size * 5;
		if (Math.fabs (user_size - best_size) > Math.fabs (min_size * 10 - best_size))
			user_size = min_size * 10;
		
		return user_size;
	}
	
	private static int[] divisor = {24, 12, 4, 3, 2, 0};
	private static int fit_to_trigonometric (double length)
	{
		for (int i = 0; divisor[i] != 0; i++) {
			if (length <= (Math.PI / divisor[i] + 1e-7))
				return divisor[i];
		}
		return -(int) Math.round (length / Math.PI);
	}
	
	private void setup_grid (Context ctx)
	{
		double unused = 0.0;
		grid_size = 70.0;
		ctx.device_to_user_distance (ref grid_size, ref unused);
		grid_size = readable_grid_size (grid_size);
	}
	
	private delegate void GridForeach (double x);
	
	private void foreach_grid_x (Context ctx, GridForeach func)
	{
		double unused = 0.0, x = 0.0;
		ctx.device_to_user (ref x, ref unused);
		double step_x = grid_size;
		if (document.trigonometric) {
			int d = fit_to_trigonometric (step_x);
			step_x = (d < 0) ? (-d * Math.PI) : (Math.PI / d);
		}
			
		for (x = x - Math.fmod (x, step_x); true; x += step_x) {
			double dx = x;
			ctx.user_to_device (ref dx, ref unused);
			if (dx > allocation.width)
				break;
			func (x);
		}
	}
	
	private void foreach_grid_y (Context ctx, GridForeach func)
	{
		double unused = 0.0, y = 0.0;
		ctx.device_to_user (ref unused, ref y);
		for (y = y - Math.fmod (y, grid_size); true; y -= grid_size) {
			double dy = y;
			ctx.user_to_device (ref unused, ref dy);
			if (dy > allocation.height)
				break;
			func (y);
		}
	}
	
	private void draw_grid (Context ctx)
	{
		ctx.save ();
		ctx.set_line_width (pixel_x * 0.2);
		ctx.set_source_rgb (0.5, 0.5, 0.5);

		double y1 = 0.0;
		double y2 = allocation.height;
		double x1 = 0.0;
		double x2 = allocation.width;
		ctx.device_to_user (ref x1, ref y1);
		ctx.device_to_user (ref x2, ref y2);
		
		double cross_size = 5 * pixel_x;

		switch (document.grid_style) {
		case GridStyle.LINES:
			foreach_grid_x (ctx, (x) => {
				ctx.move_to (x, y1);
				ctx.line_to (x, y2);
			});
			foreach_grid_y (ctx, (y) => {
				ctx.move_to (x1, y);
				ctx.line_to (x2, y);
			});
			break;
		case GridStyle.CROSSES:
			foreach_grid_x (ctx, (x) => {
				foreach_grid_y (ctx, (y) => {
					ctx.move_to (x - cross_size, y);
					ctx.line_to (x + cross_size, y);
					ctx.move_to (x, y - cross_size);
					ctx.line_to (x, y + cross_size);
				});
			});
			break;
		case GridStyle.POLAR:
			double rmin, rmax, r, a;
			rmin = rmax = Math.hypot (x1, y1);
			
			r = Math.hypot (x1, y2);
			if (r < rmin) rmin = r;
			if (r > rmax) rmax = r;
			r = Math.hypot (x2, y1);
			if (r < rmin) rmin = r;
			if (r > rmax) rmax = r;
			r = Math.hypot (x2, y2);
			if (r < rmin) rmin = r;
			if (r > rmax) rmax = r;
			r = Math.hypot (axis_x, axis_y);
			if (r < rmin) rmin = r;
			if (r > rmax) rmax = r;
			
			for (r = r - Math.fmod (rmin, grid_size); r <= rmax; r += grid_size)
				ctx.arc (0.0, 0.0, r, 0.0, Math.PI * 2);
			double PI_12 = Math.PI / 12;
			for (a = 0.0; a < Math.PI; a += PI_12) {
				if (Math.fabs (a - Math.PI_2) > 1e-7) {
					double k = Math.tan (a);
					ctx.move_to (x1, k*x1);
					ctx.line_to (x2, k*x2);
				} else {
					ctx.move_to (0.0, y1);
					ctx.line_to (0.0, y2);
				}
			}
			break;
		}
		
		ctx.stroke ();
		ctx.restore ();
	}
	
	private void setup_axes (Context ctx)
	{
		double unused = 0.0, dx, dy;
		
		axis_x = dx = 0.0;
		axis_y = dy = 0.0;
		ctx.user_to_device (ref dx, ref dy);
		
		if (dx < 5) {
			axis_x = dx = 5;
			ctx.device_to_user (ref axis_x, ref unused);
		}
		
		if (dy < 5) {
			axis_y = dy = 5;
			ctx.device_to_user (ref unused, ref axis_y);
		}
		
		if (dx > allocation.width - 5) {
			axis_x = dx = allocation.width - 5;
			ctx.device_to_user (ref axis_x, ref unused);
		}
		
		if (dy > allocation.height - 5) {
			axis_y = dy = allocation.height - 5;
			ctx.device_to_user (ref unused, ref axis_y);
		}
	}
	
	private void draw_axes (Context ctx)
	{
		double utop, ubottom, uleft, uright;

		ctx.save ();
		ctx.set_line_width (pixel_x);
		ctx.set_source_rgb (0.0, 0.0, 0.0);
		
		uleft = 0.0;
		utop = 0.0;
		ctx.device_to_user (ref uleft, ref utop);
		uright = allocation.width;
		ubottom = allocation.height;
		ctx.device_to_user (ref uright, ref ubottom);
		
		// OY axis
		ctx.move_to (axis_x, utop);
		ctx.line_to (axis_x, ubottom);
		ctx.stroke ();
		ctx.move_to (axis_x, utop);
		ctx.line_to (axis_x - pixel_x * 3, utop - (-pixel_y) * 10);
		ctx.line_to (axis_x + pixel_x * 3, utop - (-pixel_y) * 10);
		ctx.line_to (axis_x, utop);
		ctx.fill ();

		// OX axis
		ctx.move_to (uleft, axis_y);
		ctx.line_to (uright, axis_y);
		ctx.stroke ();
		ctx.move_to (uright, axis_y);
		ctx.line_to (uright - pixel_x * 10, axis_y - (-pixel_y) * 3);
		ctx.line_to (uright - pixel_x * 10, axis_y + (-pixel_y) * 3);
		ctx.line_to (uright, axis_y);
		ctx.fill ();
		
		// Draw points on axis
		ctx.set_line_width (pixel_x * 0.5);
		foreach_grid_x (ctx, (x) => {
			ctx.move_to (x, axis_y - pixel_y * 3);
			ctx.line_to (x, axis_y + pixel_y * 3);
		});

		foreach_grid_y (ctx, (y) => {
			ctx.move_to (axis_x - pixel_x * 3, y);
			ctx.line_to (axis_x + pixel_x * 3, y);
		});
		
		ctx.stroke ();
		ctx.restore ();
	}
	
	private void draw_labels (Context ctx)
	{
		Matrix matrix;
		ctx.get_matrix (out matrix);
		ctx.select_font_face ("Sans", FontSlant.NORMAL, FontWeight.NORMAL);
		
		FontExtents fe;
		ctx.identity_matrix ();
		ctx.font_extents (out fe);
		ctx.set_matrix (matrix);
		double dx = 0.0, dy = 0.0;
		ctx.user_to_device (ref dx, ref dy);
		
		double max_width = 0.0;
		foreach_grid_y (ctx, (y) => {
			if (Math.fabs (y) <= 1e-9)
				return;

			TextExtents te;
			string str_y = "%.*lf".printf (get_precision (grid_size), y);
			ctx.identity_matrix ();
			ctx.text_extents (str_y, out te);
			if (te.width > max_width)
				max_width = te.width;
			ctx.set_matrix (matrix);
		});
		int side_x = ((dx + max_width + 5) >= allocation.width) ? -1 : 1;
		foreach_grid_y (ctx, (y) => {
			if (Math.fabs (y) <= 1e-9)
				return;

			TextExtents te;
			string str_y = "%.*lf".printf (get_precision (grid_size), y);
			
			ctx.identity_matrix ();
			ctx.text_extents (str_y, out te);
			ctx.set_matrix (matrix);
			ctx.device_to_user_distance (ref te.width, ref te.height);
			ctx.device_to_user_distance (ref te.x_bearing, ref te.y_bearing);
			if (side_x > 0) {
				ctx.move_to (axis_x + pixel_x * 5.0,
				             y - te.y_bearing - te.height / 2.0);
			} else {
				ctx.move_to (axis_x - te.width - pixel_x * 5.0,
				             y - te.y_bearing - te.height / 2.0);
			}
			ctx.identity_matrix ();
			ctx.show_text (str_y);
			ctx.set_matrix (matrix);
		});

		int side_y = (dy <= (fe.height + 5)) ? -1: 1;
		bool skip = false;
		foreach_grid_x (ctx, (x) => {
			if (Math.fabs (x) <= 1e-9)
				return;
			if (skip) {
				skip = false;
				return;
			}
			
			TextExtents te;
			string str_x;
			if (document.trigonometric) {
				int a = (int) Math.fabs (Math.round (divisor[0] * x / Math.PI));
				int b = divisor[0];
				b /= gcd (a, divisor[0]);
				a /= gcd (a, divisor[0]);
				if (a != 1 && b != 1)
					str_x = "%dπ/%d".printf (a, b);
				else if (a == 1 && b != 1)
					str_x = "π/%d".printf (b);
				else if (a != 1)
					str_x = "%dπ".printf (a);
				else
					str_x = "π";
				if (x < 0) str_x = "-" + str_x;
			} else
				str_x = "%.*lf".printf (get_precision (grid_size), x);
			
			ctx.identity_matrix ();
			ctx.text_extents (str_x, out te);
			ctx.set_matrix (matrix);
			ctx.device_to_user_distance (ref te.width, ref te.height);
			ctx.device_to_user_distance (ref te.x_bearing, ref te.y_bearing);
			if (side_y > 0) {
				ctx.move_to (x - te.x_bearing - te.width / 2.0,
				             axis_y - pixel_y * 5.0);
			} else {
				ctx.move_to (x - te.x_bearing - te.width / 2.0,
				             axis_y + pixel_y * 5.0 + te.height);
			}
			ctx.identity_matrix ();
			ctx.show_text (str_x);
			ctx.set_matrix (matrix);

			if (te.width > grid_size)
				skip = true;
		});
		
		ctx.stroke ();
	}
	
	private void draw_plots (Context ctx)
	{
		document.list_foreach ( (plot) => {
			if (plot.visible) {
				ctx.save ();
				ctx.set_line_width (pixel_x * plot.line_width);
				ctx.set_line_cap (LineCap.ROUND);
				cairo_set_dash_style (ctx, plot.line_style, 0.0);
				Gdk.Color color = plot.line_color;
				ctx.set_source_rgb (color.red / 65535.0,
				                    color.green / 65535.0,
				                    color.blue / 65535.0);
				ctx.new_path ();
				plot.render (ctx, document.min_x, document.max_x,
				             0.0, 0.0, pixel_x, pixel_y);
				ctx.stroke ();
				ctx.restore ();
			}
			return false;
		});
	}
	
	private override bool expose_event (Gdk.EventExpose event)
	{
		var ctx = Gdk.cairo_create (window);
		// Clip
		Gdk.cairo_rectangle (ctx, event.area);
		ctx.clip ();
		
		// Setup transformation
		ctx.translate (-document.min_x * scale (),
		                allocation.height / 2.0 + document.shift_y);
		ctx.scale (scale (), -scale ());
		pixel_x = 1.0;
		pixel_y = 1.0;
		ctx.device_to_user_distance (ref pixel_x, ref pixel_y);
		
		setup_grid (ctx);
		draw_grid (ctx);
		setup_axes (ctx);
		if (document.show_axes)
			draw_axes (ctx);
		if (document.show_labels)
			draw_labels (ctx);
		draw_plots (ctx);
		
		return true;
	}
}
