using Gtk;

public class PlotWindow {
	/* Document */
	private Document		document;
	private bool			modified;
	/* UI */
	private Builder			ui;
	private Window			window;
	private PlotArea		plot_area;
	private TreeIter		selected_iter;
	private AbstractPlot?		selected_plot;
	/* Statusbar */
	private Statusbar		statusbar;
	private uint			status_context;
	
	public PlotWindow (File? file = null) throws Error
	{
		ui = new Builder ();
		ui.add_from_file (Path.build_filename (ui_dir, "vplot.ui"));
		window = ui.get_object ("main_window") as Window;
		document = null;
		try {
			if (file != null)
				document = new Document.from_file (file);
		} catch (MarkupError e) {
			show_error (_("Error reading file"), e.message, window);
			document = null;
		} catch (Error e) {
			show_error (_("Error opening file"), e.message, window);
			document = null;
		}
		if (document == null)
			document = new Document ();
		
		/* Setup functions view */
		var functions_view = ui.get_object ("functions_view") as TreeView;
		CellRenderer cell = new CellRendererToggle ();
		((CellRendererToggle) cell).toggled.connect( (s, path_str) => {
			TreeIter iter;
			document.get_list ().get_iter (out iter, new TreePath.from_string (path_str));
			document.list_set_visibility (iter, !document.list_get_plot (iter).visible);
		});
		functions_view.insert_column_with_attributes (-1, "", cell, "active", 0);
		cell = new CellRendererText ();
		functions_view.insert_column_with_attributes (-1, "", cell, "text", 1);
		functions_view.get_selection ().changed.connect (selection_changed);
		
		/* Connect color change signal */
		ui.get_object ("colorbutton").notify["color"].connect ((s, p) => {
			Gdk.Color color = (s as ColorButton).color;
			if (selected_plot != null && !selected_plot.line_color.equal (color))
				selected_plot.line_color = color;
		});
		
		/* Setup dash combobox and its model */
		var combobox_color = ui.get_object ("combobox_style") as ComboBox;
		var cell_dash = new CellRendererDash ();
		var dash_model = new ListStore (1, typeof (DashStyle));
		dash_model_fill (dash_model);
		combobox_color.set_model (dash_model);
		combobox_color.pack_start (cell_dash, true);
		combobox_color.add_attribute (cell_dash, "dash", 0);
		
		/* Setup statusbar */
		statusbar = ui.get_object ("statusbar") as Statusbar;
		status_context = statusbar.get_context_id ("Coordinates under the cursor");
		statusbar.push (status_context, "");
		
		ui.connect_signals (this);
		setup_document (document);
	}
	
	public PlotWindow run () {
		window.show ();
		return this;
	}
	
	private string get_document_name ()
	{
		return (document.get_file () != null) ? document.get_file ().get_basename () : _("Untitled");
	}
	
	private void make_title ()
	{
		string file_name = (modified ? "*" : "") + get_document_name ();
		window.title = "%s - %s".printf (file_name, Config.PACKAGE_NAME);
	}
	
	private void setup_document (Document newdoc)
	{
		document = newdoc;
		document.changed.connect ( () => {
			if (modified) return;
			/* Update title */
			modified = true;
			make_title ();
		});

		/* Setup plot area */
		var container = ui.get_object ("plot") as Alignment;
		if (plot_area != null) container.remove (plot_area);
		plot_area = new PlotArea (document);
		container.add (plot_area);
		(plot_area as Widget).motion_notify_event.connect (plot_motion);
		
		/* Setup UI */
		(ui.get_object ("menuitem_axes") as CheckMenuItem).active = document.show_axes;
		(ui.get_object ("menuitem_labels") as CheckMenuItem).active = document.show_labels;
		(ui.get_object ("menuitem_trigonometric") as CheckMenuItem).active = document.trigonometric;
		(ui.get_object ("functions_view") as TreeView).model = document.get_list ();

		EnumClass enum_klass = (EnumClass) typeof (GridStyle).class_ref ();
		unowned EnumValue? enum_value = enum_klass.get_value (document.grid_style);
		var obj_name = "menuitem_grid_%s".printf (enum_value.value_nick);
		(ui.get_object (obj_name) as RadioMenuItem).active = true;
		
		make_title ();
	}
	
	private void dash_model_fill (ListStore model)
	{
		TreeIter iter;
		for (DashStyle style = DashStyle.SOLID;
			 style <= DashStyle.DASH_DOT_DOT;
			 style += 1)
		{
			model.append (out iter);
			model.set (iter, 0, style, -1);
		}
	}
	
	[CCode (instance_pos = -1)]
	public void menuitem_sidebar_toggled (CheckMenuItem source)
	{
		(ui.get_object ("panel_layout") as Widget).visible = source.active;
	}

	[CCode (instance_pos = -1)]
	public void menuitem_axes_toggled (CheckMenuItem source) {document.show_axes = source.active;}
	[CCode (instance_pos = -1)]
	public void menuitem_labels_toggled (CheckMenuItem source) {document.show_labels = source.active;}
	[CCode (instance_pos = -1)]
	public void menuitem_trigonometric_toggled (CheckMenuItem source) {document.trigonometric = source.active;}	
	
	[CCode (instance_pos = -1)]
	public void menuitem_reset_view (MenuItem source)
	{
		weak ObjectClass klass = document.get_class ();
		document.set ("min_x", ((ParamSpecDouble) klass.find_property ("min_x")).default_value,
		              "max_x", ((ParamSpecDouble) klass.find_property ("max_x")).default_value,
		              "shift_y", ((ParamSpecDouble) klass.find_property ("shift_y")).default_value);
	}
	
	[CCode (instance_pos = -1)]
	public void menuitem_zoom (MenuItem source)
	{
		plot_area.zoom_to (plot_area.allocation.x + plot_area.allocation.width / 2.0,
		                   plot_area.allocation.y + plot_area.allocation.height / 2.0,
		                   (source as Buildable).get_name ().has_suffix ("in") ? 0.5 : 2.0);
	}
	
	[CCode (instance_pos = -1)]
	public void menuitem_about_activated (Menu source)
	{
		string[] authors = {
			"Sergey Nizovtsev <snizovtsev@gmail.com>"
		};
		show_about_dialog (window,
				   "authors", authors,
				   "comments", _("vPlot - Easy and handy mathematical function plotter"),
				   "copyright", "© 2009, 2010 Sergey Nizovtsev",
				   "program-name", _("Plot builder"),
				   "logo-icon-name", "vplot",
				   "translator-credits", _("translator-credits"),
				   "version", Config.PACKAGE_VERSION,
				   "website", Config.PACKAGE_URL,
				   "website-label", _("vPlot web site")
				   );
	}
	
	[CCode (instance_pos = -1)]
	public void new_plot_menu_deactivate (Menu source)
	{
		(ui.get_object ("button_add") as ToggleButton).active = false;
	}
	
	private void new_plot_menu_position_func (Menu menu, out int x, out int y, out bool push_in)
	{
		var button_add = ui.get_object ("button_add") as Button;
		button_add.window.get_position (out x, out y);
		x += button_add.allocation.x + 3;
		y += button_add.allocation.y + button_add.allocation.height + 3;
		push_in = false;
	}
	
	[CCode (instance_pos = -1)]
	public void button_add_toggled (ToggleButton source)
	{
		var new_plot_menu = ui.get_object ("new_plot_menu") as Menu;
		new_plot_menu.popup (null, null, new_plot_menu_position_func, 1, 0);
	}
	
	private void append_plot (AbstractPlot plot)
	{
		TreeIter iter;
		document.get_list ().append (out iter);
		document.list_set_plot (iter, plot);
		var functions_view = ui.get_object ("functions_view") as TreeView;
		functions_view.get_selection ().select_iter (iter);
		functions_view.scroll_to_cell (document.get_list ().get_path (iter),
		                               null, false, 0.0f, 0.0f);
	}
	
	[CCode (instance_pos = -1)]
	public void create_cartesian_plot (Menu source)
	{
		var plot = new CartesianPlot (document.get_library ());
		plot.expression = document.get_library ().suggest_free_name () + "(x) = 0";
		append_plot (plot);
		(ui.get_object ("function_entry") as Widget).grab_focus ();
	}

	[CCode (instance_pos = -1)]
	public void create_parametric_plot (Menu source)
	{
		var plot = new ParametricPlot (document.get_library ());
		string name = document.get_library ().suggest_free_name (true);
		plot.expression_x = name + "_x(t) = 0";
		plot.expression_y = name + "_y(t) = 0";
		append_plot (plot);
		(ui.get_object ("function_entry_x") as Widget).grab_focus ();
	}

	[CCode (instance_pos = -1)]
	public void create_polar_plot (Menu source)
	{
		var plot = new PolarPlot (document.get_library ());
		plot.expression = document.get_library ().suggest_free_name () + "(x) = 0";
		append_plot (plot);
		(ui.get_object ("function_entry") as Widget).grab_focus ();
	}
	
	[CCode (instance_pos = -1)]
	public void button_remove_clicked (Button source)
	{
		document.get_list ().remove (selected_iter);
	}
	
	[CCode (instance_pos = -1)]
	public void selection_changed (TreeSelection selection)
	{
		var expander1 = ui.get_object ("expander1") as Expander;
		var expander2 = ui.get_object ("expander2") as Expander;
		var expander3 = ui.get_object ("expander3") as Expander;
		var function_entry = ui.get_object ("function_entry") as Entry;
		var function_entry_x = ui.get_object ("function_entry_x") as Entry;
		var function_entry_y = ui.get_object ("function_entry_y") as Entry;
		var spinbutton_max = ui.get_object ("spinbutton_max") as SpinButton;
		var spinbutton_min = ui.get_object ("spinbutton_min") as SpinButton;
		var checkbutton_max = ui.get_object ("checkbutton_max") as CheckButton;
		var checkbutton_min = ui.get_object ("checkbutton_min") as CheckButton;
		var button_remove = ui.get_object ("button_remove") as Button;
		
		commit_expression (selected_plot);
		selected_plot = null;
		if (selection.get_selected (null, out selected_iter)) {
			var table_xy = ui.get_object ("table_xy") as Table;
			table_xy.visible = false;
			(function_entry as Widget).visible = false;
			button_remove.sensitive = true;
			expander1.sensitive = true;
			expander2.sensitive = true;
			expander3.sensitive = true;

			var plot = document.list_get_plot (selected_iter);
			(ui.get_object ("spinbutton_width") as SpinButton).value = plot.line_width;
			(ui.get_object ("colorbutton") as ColorButton).color = plot.line_color;
			(ui.get_object ("combobox_style") as ComboBox).active = plot.line_style;
			bool has_min = plot.min_arg != -double.MAX;
			bool has_max = plot.max_arg !=  double.MAX;
			if (has_min) {
				spinbutton_min.value = plot.min_arg;
				spinbutton_min.sensitive = true;
			} else {
				spinbutton_min.sensitive = false;
			}
			if (has_max) {
				spinbutton_max.value = plot.max_arg;
				spinbutton_max.sensitive = true;
			} else {
				spinbutton_max.sensitive = false;
			}
			checkbutton_min.sensitive = plot.supports_unlimited_arg ();
			checkbutton_min.active = has_min;
			checkbutton_max.sensitive = plot.supports_unlimited_arg ();
			checkbutton_max.active = has_max;
			
			if (plot is SingleExpressionPlot) {
				(function_entry as Widget).visible = true;
				function_entry.text = (plot as SingleExpressionPlot).expression;
			} else if (plot is CoupleExpressionPlot) {
				table_xy.visible = true;
				var ce_plot = plot as CoupleExpressionPlot;
				function_entry_x.text = ce_plot.expression_x;
				function_entry_y.text = ce_plot.expression_y;
			}
			
			selected_plot = plot;
		} else {
			expander1.sensitive = false;
			expander2.sensitive = false;
			expander3.sensitive = false;
			button_remove.sensitive = false;
		}
	}
	
	[CCode (instance_pos = -1)]
	public void function_entry_changed (Entry source)
	{
		document.list_set_text (selected_iter, source.text);
	}

	[CCode (instance_pos = -1)]
	public void function_entry_xy_changed (Entry source)
	{
		string text_x = (ui.get_object ("function_entry_x") as Entry).text;
		string text_y = (ui.get_object ("function_entry_y") as Entry).text;
		document.list_set_text (selected_iter, text_x + "\n" + text_y);
	}
	
	private void commit_expression (AbstractPlot? plot)
	{
		if (plot == null) return;
		
		var function_entry = ui.get_object ("function_entry") as Entry;
		var function_entry_x = ui.get_object ("function_entry_x") as Entry;
		var function_entry_y = ui.get_object ("function_entry_y") as Entry;
		
		if (plot is SingleExpressionPlot) {
			var se_plot = plot as SingleExpressionPlot;
			if (se_plot.expression != function_entry.text)
				se_plot.expression = function_entry.text;
		} else if (plot is CoupleExpressionPlot) {
			var ce_plot = plot as CoupleExpressionPlot;
			if (ce_plot.expression_x != function_entry_x.text)
				ce_plot.expression_x = function_entry_x.text;
			if (ce_plot.expression_y != function_entry_y.text)
				ce_plot.expression_y = function_entry_y.text;
		}
	}
	
	[CCode (instance_pos = -1)]
	public void function_entry_activate (Entry source)
	{
		commit_expression (selected_plot);
	}
	
	[CCode (instance_pos = -1)]
	public bool function_entry_focus_out_event (Entry source, Gdk.Event event)
	{
		commit_expression (selected_plot);
		return false;
	}
	
	[CCode (instance_pos = -1)]
	public void min_toggled (CheckButton source)
	{
		if (selected_plot == null)
			return;
		var button = ui.get_object ("spinbutton_min") as SpinButton;
		if (source.active) {
			button.sensitive = true;
			button.value = 0.0;
			selected_plot.min_arg = 0.0;
		} else {
			button.sensitive = false;
			selected_plot.min_arg = -double.MAX;
		}
	}

	[CCode (instance_pos = -1)]
	public void max_toggled (CheckButton source)
	{
		if (selected_plot == null)
			return;
		var button = ui.get_object ("spinbutton_max") as SpinButton;
		if (source.active) {
			button.sensitive = true;
			button.value = 0.0;
			selected_plot.max_arg = 0.0;
		} else {
			button.sensitive = false;
			selected_plot.max_arg = double.MAX;
		}
	}
	
	[CCode (instance_pos = -1)]
	public void min_changed (SpinButton source)
	{
		if (selected_plot == null)
			return;
		selected_plot.min_arg = source.value;
	}

	[CCode (instance_pos = -1)]
	public void max_changed (SpinButton source)
	{
		if (selected_plot == null)
			return;
		selected_plot.max_arg = source.value;
	}
	
	[CCode (instance_pos = -1)]
	public void width_changed (SpinButton source)
	{
		if (selected_plot == null)
			return;
		selected_plot.line_width = source.value;
	}
	
	[CCode (instance_pos = -1)]
	public void style_changed (ComboBox source)
	{
		if (selected_plot == null)
			return;
		selected_plot.line_style = (DashStyle) source.active;
	}
	
	[CCode (instance_pos = -1)]
	public void grid_style_changed (RadioMenuItem item)
	{
		if (item.active) {
			/* Convert menuitem_grid_%s to GridStyle */
			EnumClass klass = (EnumClass) typeof (GridStyle).class_ref ();
			unowned string nick = (item as Buildable).get_name ();
			unowned EnumValue? enum_value = klass.get_value_by_nick (nick.offset (14));
			if (document.grid_style != (GridStyle) enum_value.value)
				document.grid_style = (GridStyle) enum_value.value;
		}
	}

	private Pango.AttrList entry_attr_list;
	private uint           entry_last_hash;
	private int            entry_last_cursor_pos = -1;
	[CCode (instance_pos = -1)]
	public bool entry_highlight (Entry source, Gdk.Event event)
	{
		bool text_changed = source.text.hash () != entry_last_hash;
		bool cursor_changed = source.cursor_position != entry_last_cursor_pos;
		
		if (text_changed || cursor_changed) {
			entry_last_hash = source.text.hash ();
			entry_attr_list = new Pango.AttrList ();
			var ucolor = Pango.attr_underline_color_new (65535, 0, 0);
			var unline = Pango.attr_underline_new (Pango.Underline.ERROR);
			
			/* Underline error position */
			weak string error;
			Function? func = Function.from_string (source.text, out error);
			if (func == null) {
				int offset = (int) source.text.pointer_to_offset (error);
				ucolor.start_index = unline.start_index = source.text_index_to_layout_index (offset);
				ucolor.end_index = unline.end_index = source.text_index_to_layout_index (offset + 1);
				entry_attr_list.insert ((owned) ucolor);
				entry_attr_list.insert ((owned) unline);
			}

			/* Highlight paired brackets */
			entry_last_cursor_pos = source.cursor_position;
			unowned string start_ptr = source.text.offset (entry_last_cursor_pos);
			int open = -1, close = -1;
			
			if (start_ptr.get_char () == '(') {
				open = close = entry_last_cursor_pos;
				unowned string ptr = start_ptr.next_char ();
				int pending = 1;
				unichar c;
				
				while ((c = ptr.get_char ()) != 0 && pending > 0) {
					if (c == '(')
						pending++;
					if (c == ')')
						pending--;
					ptr = ptr.next_char ();
					close++;
				}
				if (pending != 0)
					open = -1;
			} else if (start_ptr.get_char () == ')') {
				close = open = entry_last_cursor_pos;
				unowned string ptr = start_ptr.prev_char ();
				int pending = 1;
				
				while ((void*) source.text <= (void*) ptr && pending > 0) {
					unichar c = ptr.get_char ();
					if (c == '(')
						pending--;
					if (c == ')')
						pending++;
					ptr = ptr.prev_char ();
					open--;
				}
				if (pending != 0)
					open = -1;
			}

			if (open != -1) {
				var attr_open = Pango.attr_weight_new (Pango.Weight.SEMIBOLD);
				attr_open.start_index = source.text_index_to_layout_index (open);
				attr_open.end_index = source.text_index_to_layout_index (open + 1);
				entry_attr_list.insert ((owned) attr_open);
	
				var attr_close = Pango.attr_weight_new (Pango.Weight.SEMIBOLD);
				attr_close.start_index = source.text_index_to_layout_index (close);
				attr_close.end_index = source.text_index_to_layout_index (close + 1);
				entry_attr_list.insert ((owned) attr_close);
			}
		}
		
		source.get_layout ().set_attributes (entry_attr_list);
		return false;
	}
	
	private void set_status_text (string text)
	{
		statusbar.pop (status_context);
		statusbar.push (status_context, text);
	}
	
	private bool plot_motion (Widget widget, Gdk.EventMotion event)
	{
		PlotArea plot_area = widget as PlotArea;
		double x = event.x, y = event.y;
		double ex = 1.0, ey = 1.0;
		
		plot_area.device_to_user (ref x, ref y);
		plot_area.device_to_user_distance (ref ex, ref ey);
		
		set_status_text ("x = %.*lf    y = %.*lf".printf
				 (PlotArea.get_precision (ex), x, PlotArea.get_precision (ey), y)
				);
		return false;
	}
	
	private bool do_save_as ()
	{
		var file_chooser = new FileChooserDialog (_("Save as..."),
							  window,
							  FileChooserAction.SAVE,
							  STOCK_CANCEL, ResponseType.CANCEL,
							  STOCK_SAVE, ResponseType.ACCEPT);
		file_chooser.local_only = false;
		file_chooser.do_overwrite_confirmation = true;
		file_chooser.visible = false;
		
		var response = file_chooser.run ();
		var uri = file_chooser.get_uri ();
		file_chooser.destroy ();
		
		if (response == ResponseType.ACCEPT) {
			var file = File.new_for_uri (uri);
			try {
				document.save_to_file (file);
				modified = false;
				make_title ();
				return true;
			} catch (Error e) {
				show_error (_("Could not save plot"), e.message, window);
			}
		}
		return false;
	}
	
	private bool do_save ()
	{
		if (document.get_file () != null) {
			try {
				document.save_to_file (document.get_file ());
				modified = false;
				make_title ();
				return true;
			} catch (Error e) {
				show_error (_("Could not save plot"), e.message, window);
			}
			return false;
		} else {
			return do_save_as ();
		}
	}
	
	private void do_open ()
	{
		var file_chooser = new FileChooserDialog (_("Open file"),
							  window,
							  FileChooserAction.OPEN,
							  STOCK_CANCEL, ResponseType.CANCEL,
							  STOCK_OPEN, ResponseType.ACCEPT);
		file_chooser.local_only = false;
		file_chooser.visible = false;
		
		var response = file_chooser.run ();
		var uri = file_chooser.get_uri ();
		file_chooser.destroy ();
		
		if (response == ResponseType.ACCEPT) {
			var file = File.new_for_uri (uri);
			try {
				var document = new Document.from_file (file);
				if (do_save_confirmation ()) {
					modified = false;
					setup_document (document);
				}
			} catch (MarkupError e) {
				show_error (_("Error reading file"), e.message, window);
			} catch (Error e) {
				show_error (_("Error opening file"), e.message, window);
			}
		}
	}
	
	private bool do_save_confirmation ()
	{
		if (!modified) return true;
		
		var dialog = new MessageDialog (window,
						DialogFlags.MODAL,
						MessageType.WARNING,
						ButtonsType.NONE,
						_("Save the changes to plot \n'%s' before closing?"),
						get_document_name ());
		dialog.format_secondary_text (_("If you don't save the plot, changes will be lost."));
		dialog.add_buttons (_("Do_n't save"), ResponseType.NO,
				    _("_Cancel"), ResponseType.CANCEL,
				    _("_Save"), ResponseType.YES);
		dialog.set_default_response (ResponseType.CANCEL);
		(dialog.get_image () as Image).set_from_stock (STOCK_SAVE, IconSize.DIALOG);
		dialog.title = _("Close %s").printf (get_document_name ());
		
		var response = dialog.run ();
		dialog.destroy ();
		
		switch (response) {
		case ResponseType.NO:
			return true;
		case ResponseType.CANCEL:
			return false;
		case ResponseType.DELETE_EVENT:
			return false;
		case ResponseType.YES:
			return do_save ();
		}
		assert (false);
		return true;
	}
	
	[CCode (instance_pos = -1)]
	public void menuitem_save_activated (MenuItem source) { do_save (); }
	[CCode (instance_pos = -1)]
	public void menuitem_save_as_activated (MenuItem source) { do_save_as (); }
	[CCode (instance_pos = -1)]
	public void menuitem_open_activated (MenuItem source) { do_open (); }
	[CCode (instance_pos = -1)]
	public bool delete_event (Window source, Gdk.Event event) { return ! do_save_confirmation (); }
	[CCode (instance_pos = -1)]
	public void menuitem_quit_activated (MenuItem source) {
		if (do_save_confirmation ())
			main_quit ();
	}
	
	[CCode (instance_pos = -1)]
	public void menuitem_new_activated (MenuItem source) {
		/* Vala will automatically create extra element for null-terminated arrays */
		char** argv = new string[1];
		argv[0] = prg_path;
 		try {
			Gdk.spawn_on_screen (Gdk.Screen.get_default (),
					     Environment.get_current_dir (),
					     (string[]) argv,
					     null, /* envp */
					     SpawnFlags.SEARCH_PATH, /* flags */
					     null, /* child setup */
					     null /* pid */
					     );
		} catch (Error e) {
			show_error (_("Error launching vplot"), e.message, window);
		}
		free (argv);
	}
	
	private void show_uri (string uri)
	{
		try {
			Gtk.show_uri (Gdk.Screen.get_default (), uri, Gdk.CURRENT_TIME);
		} catch (Error e) {
			show_error (_("Error opening URL"), e.message, window);
		}
	}
	
	[CCode (instance_pos = -1)]
	public void get_help (MenuItem source)
	{
		show_uri ("https://answers.launchpad.net/vplot/+addquestion");
	}

	[CCode (instance_pos = -1)]
	public void translate (MenuItem source)
	{
		show_uri ("https://translations.launchpad.net/vplot");
	}

	[CCode (instance_pos = -1)]
	public void report_bug (MenuItem source)
	{
		show_uri (Config.PACKAGE_BUGREPORT);
	}
}

